#!/usr/bin/env python
# -*- coding: utf-8 -*-

import math
import numpy as np
from scipy.special import erf, erfc

def nt(n, gm, gsd, dmin=None, dmax=10.):
    """Evaluate the total number of particles between two diameters.

    The CDF of a lognormal distribution is calculated using equation
    8.39 from Seinfeld and Pandis.

    Mathematically, it is represented as:

    .. math::

        N_t(D_p)=∫_{D_{min}}^{D_{max}}n_N(D_p^*)dD^*_p=\\frac{N_t}{2}+\\frac{N_t}{2}*erf\Big(\\frac{ln(D_p/D_{pg})}{\sqrt{2} lnσ_g}\Big) \\;\\;(cm^{-3})

    Parameters
    ----------
    n : float
        Total aerosol number concentration in units of #/cc
    gm : float
        Median particle diameter (geometric mean) in units of microns.
    gsd : float
        Geometric Standard Deviation of the distribution.
    dmin : float
        The minimum particle diameter in microns. Default value is 0 :math:`\mu m`.
    dmax : float
        The maximum particle diameter in microns. Default value is 10 :math:`\mu m`.

    Returns
    -------
    N | float
        Returns the total number of particles between dmin and dmax in units of
        [:math:`particles*cm^{-3}`]

    See Also
    --------
    opcsim.equations.pdf.dn_ddp
    opcsim.equations.pdf.dn_dlndp
    opcsim.equations.pdf.dn_dlogdp

    Examples
    --------

    Evaluate the number of particles in a simple distribution between 0 and
    2.5 :math:`\mu m`:

    >>> d = opcsim.AerosolDistribution()
    >>> d.add_mode(1e3, 100, 1.5, "mode 1")
    >>> n = opcsim.equations.cdf.nt(1e3, 0.1, 1.5, dmax=2.5)

    """
    res = (n/2.) * (1 + erf((np.log(dmax/gm)) / (np.sqrt(2) * np.log(gsd))))

    if dmin is not None and dmin > 0.0:
        res -= nt(n, gm, gsd, dmin=None, dmax=dmin)

    return res

def st(n, gm, gsd, dmin=None, dmax=10.):
    """Evaluate the total surface area of the particles between two diameters.

    The CDF of the lognormal distribution is calculated using equation 8.11
    from Seinfeld and Pandis.

    Mathematically, it is represented as:

    .. math::

        S_t=π∫_{-∞}^{∞}D_p^2n_N^e(ln D_p)d lnD_p \\;\\;(\mu m^2 cm^{-3})

    Parameters
    ----------
    n : float
        Total number of particles in units of #/cc
    gm : float
        Median particle diameter (geometric mean) in units of :math:`\mu m`.
    gsd : float
        Geometric Standard Deviation of the distribution.
    dmin : float
        The minimum particle diameter in microns. Default value is 0 :math:`\mu m`.
    dmax : float
        The maximum particle diameter in microns. Default value is 10 :math:`\mu m`.

    Returns
    -------
    Surface Area | float
        Returns the total surface area of particles between :math:`D_{min}`
        and :math:`D_{max}` in units of :math:`\mu m^2 \\; cm^{-3}`

    See Also
    --------
    opcsim.equations.pdf.ds_ddp
    opcsim.equations.pdf.ds_dlndp
    opcsim.equations.pdf.ds_dlogdp

    Examples
    --------

    Integrate a sample distribution between 0 and 2.5 microns:

    >>> d = opcsim.AerosolDistribution()
    >>> d.add_mode(1e3, 100, 1.5, "mode 1")
    >>> n = opcsim.equations.cdf.st(1e3, 0.1, 1.5, dmax=2.5)

    """
    res = (np.pi/2.)*n*(gm**2) * np.exp(2*(np.log(gsd)** 2)) * \
                erfc((np.sqrt(2) * np.log(gsd)) - (np.log(dmax/gm) / (np.sqrt(2) * np.log(gsd))))

    if dmin is not None and dmin > 0.0:
        res -= st(n, gm, gsd, dmin=None, dmax=dmin)

    return res

def vt(n, gm, gsd, dmin=None, dmax=10.):
    """Evaluate the total volume of the particles between two diameters.

    The CDF of the lognormal distribution is calculated using equation 8.12
    from Seinfeld and Pandis.

    Mathematically, it is represented as:

    .. math::

        V_t=\\frac{π}{6}∫_{-∞}^{∞}D_p^3n_N^e(ln D_p)d lnD_p \\;\\;(\mu m^3 cm^{-3})

    Parameters
    ----------
    n : float
        Total aerosol number concentration in units of #/cc
    gm : float
        Median particle diameter (geometric mean) in units of :math:`\mu m`.
    gsd : float
        Geometric Standard Deviation of the distribution.
    dmin : float
        The minimum particle diameter in microns. Default value is 0 :math:`\mu m`.
    dmax : float
        The maximum particle diameter in microns. Default value is 10 :math:`\mu m`.

    Returns
    -------
    Volume | float
        Returns the total volume of particles between :math:`D_{min}` and
        :math:`D_{max}` in units of :math:`\mu m^3 cm^{-3}`

    See Also
    --------
    opcsim.equations.pdf.dv_ddp
    opcsim.equations.pdf.dv_dlndp
    opcsim.equations.pdf.dv_dlogdp

    Examples
    --------

    Integrate a sample distribution between 0 and 2.5 microns:

    >>> d = opcsim.AerosolDistribution()
    >>> d.add_mode(1e3, 100, 1.5, "mode 1")
    >>> n = opcsim.equations.cdf.vt(1e3, 0.1, 1.5, dmax=2.5)

    """

    res = (np.pi/12.)*n*(gm**3) * np.exp(9./2.*(np.log(gsd)**2)) * \
                erfc((1.5*np.sqrt(2) * np.log(gsd)) - (np.log(dmax/gm) / (np.sqrt(2) * np.log(gsd))))

    if dmin is not None and dmin > 0.0:
        res -= vt(n, gm, gsd, dmin=None, dmax=dmin)

    return res



from pkg_resources import get_distribution

import warnings
import pandas as pd
import numpy as np
import math
import os

from .distributions import *
from .models import *
from .plots import *
from .utils import *
from .metrics import *
from .mie import *
from .rc_style import set

set()

__version__ = get_distribution("opcsim").version


# -*- coding: utf-8 -*-
"""
"""
from .equations.pdf import *
from .equations.cdf import *
from .utils import k_kohler, rho_eff, RHO_H20

DISTRIBUTION_DATA = {
    'urban': [
        (7100, 0.0117, 0.232, "Mode I"),
        (6320, 0.0373, 0.250, "Mode II"),
        (960, 0.151, 0.204, "Mode III")
        ],
    'marine': [
        (133, 0.008, 0.657, "Mode I"),
        (66.6, 0.266, 0.210, "Mode II"),
        (3.1, 0.58, 0.396, "Mode III")
        ],
    'rural': [
        (6650, 0.015, 0.225, "Mode I"),
        (147, 0.054, 0.557, "Mode II"),
        (1990, 0.084, 0.266, "Mode III")
        ],
    'remote continental': [
        (3200, 0.02, 0.161, "Mode I"),
        (2900, 0.116, 0.217, "Mode II"),
        (0.3, 1.8, 0.380, "Mode III")
        ],
    'free troposphere': [
        (129, 0.007, 0.645, "Mode I"),
        (59.7, 0.250, 0.253, "Mode II"),
        (63.5, 0.52, 0.425, "Mode III")
        ],
    'polar': [
        (21.7, 0.138, 0.245, "Mode I"),
        (0.186, 0.75, 0.300, "Mode II"),
        (3e-4, 8.6, 0.291, "Mode III")
        ],
    'desert': [
        (726, 0.002, 0.247, "Mode I"),
        (114, 0.038, 0.770, "Mode II"),
        (0.178, 21.6, 0.438, "Mode III")
        ],
}

def _get_pdf_func(base, weight, dp, n, gm, gsd, rho=1.):
    """"""
    weight  = weight.lower()

    if weight not in ['number', 'surface', 'volume', 'mass']:
        raise Exception("Invalid argument for weight: ['number', 'surface', 'volume', 'mass']")

    if base not in [None, 'none', 'log', 'log10']:
        raise Exception("Invalid argument for base: ['none', 'log', 'log10']")

    if base == 'none' or base == None:
        if weight == 'number':
            return dn_ddp(dp, n, gm, gsd)
        elif weight == 'surface':
            return ds_ddp(dp, n, gm, gsd)
        elif weight == 'volume':
            return dv_ddp(dp, n, gm, gsd)
        elif weight == 'mass':
            return dv_ddp(dp, n, gm, gsd) * rho * 1e-6
    elif base == 'log':
        if weight == 'number':
            return dn_dlndp(dp, n, gm, gsd)
        elif weight == 'surface':
            return ds_dlndp(dp, n, gm, gsd)
        elif weight == 'volume':
            return dv_dlndp(dp, n, gm, gsd)
        elif weight == 'mass':
            return dv_dlndp(dp, n, gm, gsd) * rho *1e-6
    elif base == 'log10':
        if weight == 'number':
            return dn_dlogdp(dp, n, gm, gsd)
        elif weight == 'surface':
            return ds_dlogdp(dp, n, gm, gsd)
        elif weight == 'volume':
            return dv_dlogdp(dp, n, gm, gsd)
        elif weight == 'mass':
            return dv_dlogdp(dp, n, gm, gsd) * rho * 1e-6


def _get_cdf_func(n, gm, gsd, dmin=None, dmax=10., weight='number', rho=1.):
    """"""
    weight = weight.lower()
    if weight not in ['number', 'surface', 'volume', 'mass']:
        raise Exception("Invalid argument for weight: ['number', 'surface', 'volume', 'mass']")

    if weight == 'number':
        return nt(n, gm, gsd, dmin, dmax)
    elif weight == 'surface':
        return st(n, gm, gsd, dmin, dmax)
    elif weight == 'volume':
        return vt(n, gm, gsd, dmin, dmax)
    elif weight == 'mass':
        return vt(n, gm, gsd, dmin, dmax) * rho


def load_distribution(label):
    """Load sample distributions as described by Seinfeld+Pandis Table 8.3.

    There are currently 7 options including: Urban, Marine, Rural, Remote
    continental, Free troposphere, Polar, and Desert.

    Parameters
    ----------

    label : {'Urban' | 'Marine' | 'Rural' | 'Remote Continental' | 'Free Troposphere' | 'Polar' | 'Desert'}
        Choose which sample distribution to load.

    Returns
    -------
    An instance of the AerosolDistribution class

    Examples
    --------

    >>> d = opcsim.load_distribution("Urban")

    """
    label = label.lower()

    if label not in DISTRIBUTION_DATA.keys():
        raise ValueError("Invalid label.")

    _tmp = AerosolDistribution(label)

    for each in DISTRIBUTION_DATA[label]:
        _tmp.add_mode(each[0], each[1], 10**each[2], each[3])

    return _tmp


class AerosolDistribution(object):
    """Define an aerosol distribution.

    Assuming an aerosol distribution can be described as the sum of n lognormal
    modes, define as (Seinfeld and Pandis equation 8.54):

    .. math::

        n_N^o(logD_p)=Σ_{i=1}^{n}\\frac{N_i}{\sqrt{2π} logσ_i}exp(-\\frac{(logD_p - logD̄_{pi})^2}{2log^2σ_i})

    """
    def __init__(self, label=None):
        """Initialize an Aerosol Distribution.

        Parameters
        ----------
        label : string, optional
            Label the distribution

        Returns
        -------
        AerosolDistribution
            An instance of the AerosolDistribution class

        Examples
        --------

        >>> d = AerosolDistribution("Urban")

        """
        self.label = label
        self.modes = []

    def _get_mode(self, label):
        """Return the mode with label=`label`. """
        for mode in self.modes:
            if mode['label'].lower() == label.lower():
                return mode

        return None

    def add_mode(self, n, gm, gsd, label=None, kappa=0., rho=1., refr=complex(1.5, 0)):
        """Add a mode to the distribution as defined using N, GM, and GSD. Additionally,
        each mode has optical and chemical properties that can be set including: the k-kohler 
        coefficient, the density, and the refractive index.

        Parameters
        ----------
        n : float
            Total number of particles (#/cc)
        gm : float
            Median particle diameter (Geometric Mean) in units of microns.
        gsd : float
            Geometric Standard Deviation
        label : string, optional
            Label for the mode
        kappa: float, optional
            The k-kohler coefficient to describe hygroscopic growth. Default is 0.
        rho: float: optional
            The particle density in units of g/cm3. Default is 1.
        refr: complex: optional
            The complex refractive index. Default is 1.5+0i. This should be the dry 
            refractive index. If/when the distribution is evaluated at a non-zero RH,
            the refractive index will be adjusted based on it's water content at that 
            RH using the defined kappa value.

        Examples
        --------

        Create a single-mode aerosol distribution with parameters N=7100,
        GM=11.7 nm, GSD=1.706. These are the sample parameters for the first
        mode of a typical Urban aerosol distribution as described in Seinfeld
        and Pandis (Table 8.3).

        >>> d = AerosolDistribution("Urban")
        >>> d.add_mode(n=7100, gm=0.0117, gsd=1.706, label="Mode 1")

        We can also go ahead and add the second and third modes if we so choose:

        >>> d.add_mode(n=6320, gm=0.0373, gsd=1.778, label="Mode 2")
        >>> d.add_mode(n=960, gm=0.151, gsd=1.599, label="Mode 3")

        """
        self.modes.append(
            {
                'label':label if label else "Mode {}".format(len(self.modes)), 
                'N':n, 'GM':gm, 
                'GSD':gsd, 
                "kappa": kappa, 
                "rho": rho, 
                "refr": refr
                }
            )

    def pdf(self, dp, base='log10', weight='number', mode=None, rh=0., rho=None):
        """Evaluate and return the probability distribution function at
        particle diameter `dp`.

        Using equation 8.54 from Seinfeld and Pandis, we can evaluate the
        probability distribution function for a multi-modal aerosol
        distribution by summing the individual pdf's. By default, the calculations
        are based on the dry diameter of the particle (assuming RH=0); however,
        you can adjust the relative humidity using the **rh** parameter which 
        will calculate the size based on growth per k-kohler theory using the 
        k-kohler coefficient set for each individual mode.

        Parameters
        ----------
        dp : float or an array of floats
            Particle diameter(s) to evaluate the pdf (um)
        base : {None | 'none' | 'log' | 'log10'}
            Base algorithm to use. Default is 'log10'
        weight : {'number' | 'surface' | 'volume' | 'mass'}
            Choose how to weight the pdf. Default is `number`.
        mode : string or None
            Choose to only evaluate the pdf for a single mode
            of the entire distribution. If set to `None`, the entire
            distribution will be evaluated.
        rh: float: optional
            The relative humidity as a percentage (0-100). Default is 0.
        rho: float: optional
            The particle density. If set, this will override the density set 
            for individual modes.

        Returns
        -------
        float
            The evaluated pdf at particle diameter dp

        Examples
        --------

        Evaluate the PDF at 100 nm

        >>> d = opcsim.load_distribution("Urban")
        >>> d.pdf(0.1)

        Evaluate the PDF at a range of particle diameters

        >>> d.pdf(np.linespace(0.1, 1., 100))

        Evaluate the PDF for a volume-weighted distribution -> returns dVdlogDp

        >>> d.pdf(0.1, weight='volume')

        Evaluate the PDF for a volume-weighted distribution in ln-space

        >>> d.pdf(0.1, weight='volume', base='log')

        """
        value = 0.0

        if mode is not None:
            modes = [self._get_mode(mode)]
        else:
            modes = self.modes

        for mode in modes:
            # calculate the wet diameter of the particle based on k-kohler theory
            gm = k_kohler(diam_dry=mode['GM'], kappa=mode["kappa"], rh=rh)

            # override rho if set 
            rho = rho if rho else mode["rho"]

            # calculate the effective density, taking into account hygroscopic growth
            rho = rho_eff([rho, RHO_H20], diams=[mode['GM'], gm - mode['GM']])

            value += _get_pdf_func(base, weight, dp, mode['N'], gm, mode['GSD'], rho)

        return value

    def cdf(self, dmax, dmin=None, weight='number', mode=None, rh=0., rho=None):
        """Evaluate and return the cumulative probability distribution function
        between `dmin` and `dmax`.

        Using equation _ from Seinfeld and Pandis, we can evaluate the cdf of a
        a multi-modal particle size distribution by summing the individual
        distributions. For example, evaluating the cdf over the entire size
        range will return the total number of particles in that size range. If
        weighted by surface area or volume, it will return the integrated
        surface area or volume respectively. By default, the calculations
        are based on the dry diameter of the particle (assuming RH=0); however,
        you can adjust the relative humidity using the **rh** parameter which 
        will calculate the size based on growth per k-kohler theory using the 
        k-kohler coefficient set for each individual mode.

        Parameters
        ----------
        dmin : float
            The minimum particle diameter in the integration (um)
        dmax : float
            The maximum particle diameter in the integration (um)
        weight : {'number' | 'surface' | 'volume' | 'mass'}
            Choose how to weight the pdf. Default is `number`
        mode : string or None
            Choose to only evaluate the pdf for a single mode
            of the entire distribution. If set to `None`, the entire
            distribution will be evaluated.
        rh: float: optional
            The relative humidity as a percentage (0-100). Default is 0.
        rho: float: optional
            The particle density. If set, this will override the density set 
            for individual modes.

        Returns
        -------
        float
            The integrated distribution function representing the total number of
            {particles, surface area, volume} between dmin and dmax.

        Examples
        --------

        Evaluate the CDF between over the entire distribution:

        >>> d = opcsim.load_distribution("Urban")
        >>> d.cdf()

        Evaluate the CDF up to 100 nm

        >>> d.cdf(dmax=0.1)

        Evaluate the total volume/mass under 2.5 microns

        >>> d.cdf(dmax=2.5, weight='volume')

        """
        if dmin is not None:
            if not dmin < dmax:
                raise ValueError("dmin must be less than dmax")

        value = 0.0

        if mode is not None:
            modes = [self._get_mode(mode)]
        else:
            modes = self.modes

        for mode in modes:
            # calculate the wet diameter of the particle based on k-kohler theory
            gm = k_kohler(diam_dry=mode['GM'], kappa=mode["kappa"], rh=rh)

            # override rho if set
            rho = rho if rho else mode["rho"]

            # calculate the effective density, taking into account hygroscopic growth
            rho = rho_eff([rho, RHO_H20], diams=[mode['GM'], gm - mode['GM']])

            value += _get_cdf_func(mode['N'], gm, mode['GSD'], dmin, dmax, weight, rho)

        return value

    def __repr__(self):
        return "AerosolDistribution: {}".format(self.label)



import numpy as np
import math
import scipy.special as bessel


def coef_pi_tau(theta, x):
    """Compute the angle-dependant functions (pi and tau) using upward recurrence.

    Per Bohren and Huffman (1983) eq. 4.47, the angle-dependant functions :math:`\pi_n` and 
    :math:`\\tau_n` can be computed using upward recurrance from the relations:

    .. math::

        \pi_n=\\frac{2n-1}{n-1}\mu\pi_{n-1}-\\frac{n}{n-1}\pi_{n-2} 
        
    .. math::
        
        \\tau_n=n\mu\pi_n - (n+1)\pi_{n-1}

    

    Parameters
    ----------
    theta: float
        The scattering angle in degrees.
    x: float
        The dimensionless size parameter, used to determine the number of elements to compute.

    Returns
    -------
    `\pi_n`, `\\tau_n`: np.ndarray of floats

    """
    # compute mu = cos(theta)
    mu = np.cos(math.radians(theta))

    # compute the max number of iterations
    nc = int(np.round(2 + x + 4*np.power(x, 1./3.)))

    # init arrays to hold the values
    pi, tau = np.zeros(nc), np.zeros(nc)

    # set the initial params
    pi[0] = 1
    pi[1] = 3*mu
    tau[0] = mu
    tau[1] = 3*np.cos(2*np.arccos(mu))

    # iterate and solve
    for n in range(2, nc):
        pi[n] = (mu*pi[n-1]*(2*n+1) - (pi[n-2]*(n+1)))/n
        tau[n] = (n+1)*mu*pi[n] - (n+2)*pi[n-1]

    return pi, tau


def coef_ab(refr, x):
    """Compute the external field coefficients using the logarithmic derivative.

    Bohren and Huffman (1983) equations 4.88 and 4.89 show how to compute :math:`a_n` and :math:`b_n`
    using the logarithmic derivative (Aden, 1951).

    .. math::

        a_n=\\frac{[D_n(mx)/m + n/x]\psi_n(x) - \psi_{n-1}(x)}{[D_n(mx)/m + n/x]\\xi_n(x) - \\xi_{n-1}(x)},

    .. math::

        b_n=\\frac{[mD_n(mx) + n/x]\psi_n(x) - \psi_{n-1}(x)}{[mD_n(mx) + n/x]\\xi_n(x) - \\xi_{n-1}(x)}

    where the logarithmic derivative is computed as:

    .. math::

        D_{n-1}=\\frac{n}{\\rho}-\\frac{1}{D_n + n/\\rho}

    
    Parameters
    ----------
    refr: complex
        The complex refractive index of the material.
    x: float
        The dimensionless size parameter.

    Returns
    -------
    `a_n`, `b_n`: np.ndarray of floats
        The external field coefficients.

    """
    # compute the number of values to calculate
    nc = int(np.round(2 + x + 4*np.power(x, 1/3)))

    # calculate z, the product of the RI and dimensionless size parameter
    z = refr*x

    nmx = int(np.round(max(nc, np.abs(z)) + 16))

    #
    n = np.arange(1, nc + 1)
    nu = n + 0.5

    # use scipy's bessel functions to compute
    sqx = np.sqrt(0.5 * np.pi * x)

    px = sqx * bessel.jv(nu, x)
    p1x = np.append(np.sin(x), px[0:nc-1])

    chx = -sqx*bessel.yv(nu, x)
    ch1x = np.append(np.cos(x), chx[0:nc-1])

    gsx = px - (0 + 1j)*chx
    gs1x = p1x - (0 + 1j)*ch1x

    # Bohren & Huffman eq. 4.89
    dn = np.zeros(nmx, dtype=np.complex128)

    for i in range(nmx-1, 1, -1):
        dn[i-1] = (i/z) - (1 / (dn[i] + i/z))

    # drop terms beyond nc
    d = dn[1:nc+1]

    da = d/refr + n/x
    db = refr*d + n/x

    an = (da*px - p1x) / (da*gsx - gs1x)
    bn = (db*px - p1x) / (db*gsx - gs1x)

    return an, bn


def s1s2(refr, x, theta):
    """Compute the complex scattering amplitudes S1 and S2 at angle theta.

    Bohren and Huffman (1983) list the equations for computing the complex scattering 
    amplitudes as Eq. 4.74:

    .. math::

        S_1=\sum_{n=1}^{n_c}\\frac{2n+1}{n(n+1)}(a_n\pi_n + b_n\\tau_n),

    .. math::

        S_2=\sum_{n=1}^{n_c}\\frac{2n+1}{n(n+1)}(a_n\\tau_n + b_n\pi_n)

    Parameters
    ----------
    refr: complex
        The complex refractive index.
    x: float
        The dimensionless size parameter.
    theta: float
        The scattering angle in degrees.

    Returns
    -------
    `S_1`, `S_2`: complex
        The complex scattering amplitudes.

    """
    # compute the number of coefficients to calculate
    nc = int(np.round(2 + x + 4*np.power(x, 1/3)))

    # compute the external field coefficients
    an, bn = coef_ab(refr=refr, x=x)

    # compute the pi and tau coefficients
    pi, tau = coef_pi_tau(theta=theta, x=x)

    # init an array for holding the S1 and S2 values
    n = np.arange(1, nc+1)
    
    # compute the coef for the series
    cn = (2*n + 1) / (n*(n+1))

    # compute the series and sum
    S1 = (cn * (an*pi + bn*tau)).sum()
    S2 = (cn * (an*tau + bn*pi)).sum()

    return S1, S2


def cscat(dp, wl, refr, theta1, theta2, nsteps=100, **kwargs):
    """Compute the scattering cross section between two angles according to Jaenicke and Hanusch (1993).

    Following the lead of Jaenicke and Hanusch (1993), we can compute the scattering cross section for 
    a given viewing angle [:math:`\Theta_1` - :math:`\Theta_2`] as:

    .. math::

        C_{sca}=\\frac{\lambda^2}{4\pi} \int_{\Theta_1}^{\Theta_2}[i_1(\Theta) + i_2(\Theta)]sin\Theta d\Theta

    where :math:`\lambda` is the incident wavelength of light and :math:`i_1` and :math:`i_2` are the intensity 
    distribution functions, calculated as:

    .. math::

        i_1(\Theta)=\mid S_1(\Theta) \mid^2,

    .. math::

        i_2(\Theta)=\mid S_2(\Theta) \mid^2


    The integral is calculated step-wise using the numpy.trapz function.
    

    Parameters
    ----------
    dp: float
        The particle diameter in microns.
    wl: float
        The wavelength of incident light in microns.
    refr: complex
        The complex refractive index of the material.
    theta1: float
        The angle from which to begin the integration.
    theta1: float
        The angle from which to end the integration.
    nsteps: int
        The number of steps in theta to use in performing the step-wise integration.
    
    Returns
    -------
    `C_{scat}`: float
        The scattering cross-section.

    """
    # build an array of angles
    thetas = np.linspace(theta1, theta2, nsteps)

    # compute the dimensionless parameter x
    x = dp*np.pi / wl

    # init an array to hold the inside of the integral
    rv = np.zeros(nsteps)
    
    # iterate over each step and compute the inside part of the integral
    for i in range(nsteps):
        s1, s2 = s1s2(refr=refr, x=x, theta=thetas[i])

        # compute i1 and i2
        i1 = (s1 * np.conjugate(s1)).real
        i2 = (s2 * np.conjugate(s2)).real

        rv[i] = (i1 + i2) * np.sin(math.radians(thetas[i]))
    
    # convert the array of angles from degrees to radians
    thetas = np.array([math.radians(x) for x in thetas])

    # compute cscat (convert the wavelength to cm to make units match common literature values)
    rv = ((wl*1e-4)**2 / (4*np.pi)) * np.trapz(rv, thetas)

    return rv



import seaborn as sns
import numpy as np
import matplotlib.pyplot as plt
import matplotlib.ticker as mtick
import itertools

from .distributions import AerosolDistribution
from .models import OPC
from .mie import cscat

lrg_number_fmt = mtick.ScalarFormatter()
lrg_number_fmt.set_powerlimits((-3, 4))

rc_log = {
    'xtick.major.size': 10.0,
    'xtick.minor.size': 6.0,
    'ytick.major.size': 10.0,
    'ytick.minor.size': 8.0,
    'xtick.color': '0.1',
    'ytick.color': '0.1',
    'axes.linewidth': 1.75
}

YLABEL = {
    'none': {
        'number': "$dN/dD_p \; [\mu m^{-1} cm^{-3}]$",
        'surface': "$dS/dD_p \; [\mu m cm^{-3}]$",
        'volume': "$dV/dD_p \; [\mu m^2 cm^{-3}]$",
        'mass': "$dM/dD_p$"
    },
    'log': {
        'number': "$dN/dlnD_p \; [# cm^{-3}]$",
        'surface': "$dS/dlnD_p \; [\mu m^2 cm^{-3}]$",
        'volume': "$dV/dlnD_p \; [\mu m^3 cm^{-3}]$",
        'mass': "$dM/dlnD_p \; [\mu g m^{-3}]$"
    },
    'log10': {
        'number': "$dN/dlogD_p \; [cm^{-3}]$",
        'surface': "$dS/dlogD_p \; [\mu m^2 cm^{-3}]$",
        'volume': "$dV/dlogD_p \; [\mu m^3 cm^{-3}]$",
        'mass': "$dM/dlogD_p \; [\mu g m^{-3}]$"
    }
}

YLABEL_CDF = {
    'number': 'Total Number [$cm^{-3}$]',
    'surface': 'Total Surface Area [$\mu m^2cm^{-3}$]',
    'volume': 'Total Volume [$\mu m^3 cm^{-3}$]',
    'mass': "Total Mass [$\mu g m^{-3}$]"
}

def histplot(data, bins, ax=None, plot_kws={}, fig_kws={}, **kwargs):
    """Plot the particle size distribution as a histogram/bar chart.

    Parameters
    ----------
    data : array of floats
        An array containing the y variable you are plotting (ex. dNdlogDp)
    bins : 3xn array
        A 3xn array containing the bin values for an OPC
    ax : matplotlib axis
        If an axis is provided, the histogram will be plotted on this axis.
        Otherwise, a new axis object will be created.
    plot_kws : dict
        Optional keyword arguments to include. They are sent as an argument to
        the matplotlib bar plot.
    fig_kws : dict
        Optional keyword arguments to include for the figure.

    Returns
    -------
    ax : matplotlib axis object

    Examples
    --------

    Plot a 10-bin OPC's response to the Urban Distribution

    .. plot::
        :context: close-figs

        >>> import opcsim, seaborn as sns
        >>> opc = opcsim.OPC(wl=0.658, n_bins=10, dmin=0.3)
        >>> opc.calibrate("psl")
        >>> d = opcsim.load_distribution("Urban")
        >>> ax = opcsim.plots.histplot(opc.evaluate(d), opc.bins)
        >>> ax.set_ylabel("$dN/dlogD_p$")
        >>> sns.despine()

    We can also plot the same OPC in volume (mass) space

    .. plot::
        :context: close-figs

        >>> ax = opcsim.plots.histplot(
        ...         opc.evaluate(d, weight='volume'), opc.bins)
        >>> ax.set_ylabel("$dV/dlogD_p$")
        >>> sns.despine()

    How about overlaying two OPC's

    .. plot::
        :context: close-figs

        >>> opcb = opcsim.OPC(wl=0.658, n_bins=5, dmin=0.3)
        >>> opcb.calibrate("psl")
        >>> ax = opcsim.plots.histplot(opc.evaluate(d),
        ...             opc.bins, label="10 bin OPC")
        >>> ax = opcsim.plots.histplot(opcb.evaluate(d), opcb.bins,
        ...             label="5 bin OPC", ax=ax)
        >>> ax.set_ylabel("$dN/dlogD_p$")
        >>> ax.legend(loc='best')
        >>> sns.despine()


    What if we want to fill in the boxes?

    .. plot::
        :context: close-figs

        >>> plot_kws = dict(fill=True)
        >>> ax = opcsim.plots.histplot(opc.evaluate(d),
        ...             opc.bins, plot_kws=plot_kws)
        >>> ax.set_ylabel("$dN/dlogD_p$")
        >>> sns.despine()

    """
    # Set the default figure kws
    default_fig_kws = dict()
    fig_kws = dict(default_fig_kws, **fig_kws)

    # Make a new axis object if one wasnt' set
    if ax is None:
        plt.figure(**fig_kws)
        ax = plt.gca()

    # Get the next color available in the palette
    nc = next(ax._get_lines.prop_cycler)['color']

    # Set the plot_kws as a mapping of default and kwargs
    default_plot_kws = dict(
                        alpha=1,
                        edgecolor=nc,
                        color=nc,
                        linewidth=5,
                        fill=False,
                        label=kwargs.pop('label', None))

    # Set the plot_kws
    plot_kws = dict(default_plot_kws, **plot_kws)

    # Plot the bar plot
    ax.bar(x=bins[:, 0], height=data, width=bins[:, -1] - bins[:, 0],
            align='edge', **plot_kws)

    # Set the xaxis to be log10
    ax.semilogx()

    # Set the xlabel
    ax.set_xlabel("$D_p \; [\mu m]$")

    ax.xaxis.set_major_formatter(mtick.FormatStrFormatter("%.3g"))

    return ax


def pdfplot(distribution, ax=None, weight='number', base='log10', with_modes=False,
            fill=False, plot_kws={}, fig_kws={}, fill_kws={}, **kwargs):
    """Plot the PDF of an aerosol size distribution.

    Parameters
    ----------
    distribution : valid `AerosolDistribution`
        An aerosol distribution with the method `pdf` that can be evaluated at
        an array of particle diameters.
    weight : {'number' | 'surface' | 'volume' | 'mass'}
        Choose how to weight the pdf. Default is `number`.
    base : {'none' | 'log' | 'log10'}
        Base algorithm to use. Default is 'log10'.
    ax : matplotlib axis
        If an axis is provided, the histogram will be plotted on this axis.
        Otherwise, a new axis object will be created.
    with_modes : bool
        If true, all modes of a given distribution will be plotted along with
        their sum. If false, only the sum will be plotted.
    fill : bool
        If true, the area under the PDF will be filled. Cannot be used while
        plotting individual modes (with_modes=True).
    plot_kws : dict
        Optional keyword arguments to include. They are sent as an argument to
        the matplotlib plot call.
    fig_kws : dict
        Optional keyword arguments to include for the figure.
    fill_kws : dict
        Optional keyword arguments to include for the fill. Sent to
        matplotlib.axes.fill_between.

    Returns
    -------
    ax : matplotlib axis object

    Examples
    --------

    Plot the number-weighted Urban aerosol distribution

    .. plot::
        :context: close-figs

        >>> import opcsim, seaborn as sns
        >>> d = opcsim.load_distribution("Urban")
        >>> ax = opcsim.plots.pdfplot(d)
        >>> ax.set_title("Urban Aerosol Distribution", fontsize=16)
        >>> sns.despine()

    Let's take a look at the individual modes as well

    .. plot::
        :context: close-figs

        >>> ax = opcsim.plots.pdfplot(d, with_modes=True)
        >>> ax.set_title("Urban Aerosol Distribution", fontsize=16)
        >>> ax.legend(loc='best')
        >>> sns.despine()

    Let's plot the volume weighted version

    .. plot::
        :context: close-figs

        >>> ax = opcsim.plots.pdfplot(d, weight='volume', with_modes=True)
        >>> ax.set_title("Volume Weighted Urban Aerosol Distribution", fontsize=16)
        >>> ax.legend(loc='best')
        >>> sns.despine()

    Let's plot a few different distributions together

    .. plot::
        :context: close-figs

        >>> d2 = opcsim.load_distribution("Marine")
        >>> d3 = opcsim.load_distribution("Rural")
        >>> ax = opcsim.plots.pdfplot(d)
        >>> ax = opcsim.plots.pdfplot(d2, ax=ax)
        >>> ax = opcsim.plots.pdfplot(d3, ax=ax)
        >>> ax.set_title("Various Aerosol Distributions", fontsize=16)
        >>> ax.legend(loc='best')
        >>> sns.despine()

    """
    if not hasattr(distribution, 'pdf'):
        raise Exception("Invalid AerosolDistribution.")

    if weight not in ['number', 'surface', 'volume', 'mass']:
        raise ValueError("Invalid weight: ['number', 'surface', 'volume', 'mass']")

    # Set the default dp values to plot against
    dp = kwargs.pop('dp', np.logspace(-3, 1, 1000))

    # Set the default figure kws
    default_fig_kws = dict()
    fig_kws = dict(default_fig_kws, **fig_kws)

    # Make a new axis object if one wasnt' set
    if ax is None:
        plt.figure(**fig_kws)
        ax = plt.gca()

    # Plot the compete distribution
    nc = next(ax._get_lines.prop_cycler)['color']

    # Set the plot_kws as a mapping of default and kwargs
    default_plot_kws = dict(
                        alpha=1,
                        linewidth=4,
                        color=nc
                        )

    # Set the plot_kws
    plot_kws = dict(default_plot_kws, **plot_kws)

    # Set the default fill_kws
    default_fill_kws = dict(color=plot_kws['color'])
    fill_kws = dict(default_fill_kws, **fill_kws)

    # If label kwarg is present, use -> otherwise use default
    label = kwargs.pop('label', distribution.label)

    # Get the data to plot
    data = distribution.pdf(dp, base=base, weight=weight, **kwargs)

    # If fill is selected, fill the gap, otherwise just plot a line
    ax.plot(dp, data, label=label, **plot_kws)

    if fill:
        ax.fill_between(dp, 0, data, label=label, **fill_kws)

    # Get data
    if with_modes is True:
        for m in distribution.modes:
            data = distribution.pdf(dp, base=base, weight=weight, mode=m['label'])

            # Pop off color from the plot_kws
            plot_kws['color'] = next(ax._get_lines.prop_cycler)['color']

            ax.plot(dp, data, label=m['label'], ls='--', **plot_kws)

    if base is not ('none' or None):
        ax.semilogx()
        ax.xaxis.set_major_formatter(mtick.FormatStrFormatter("%.4g"))

    ax.set_xlabel("$D_p \; [\mu m]$")
    ax.set_ylabel(YLABEL[base][weight])

    # Set the yaxis to show scientific notation when numbers > 1e4
    ax.yaxis.set_major_formatter(lrg_number_fmt)

    return ax


def cdfplot(distribution, ax=None, weight='number', plot_kws={},
            fig_kws={}, **kwargs):
    """Plot the CDF of a particle size distribution.

    Parameters
    ----------
    distribution : valid `AerosolDistribution`
        An aerosol distribution with the method `pdf` that can be evaluated at
        an array of particle diameters.
    weight : {'number' | 'surface' | 'volume' | 'mass'}
        Choose how to weight the pdf. Default is `number`.
    ax : matplotlib axis
        If an axis is provided, the histogram will be plotted on this axis.
        Otherwise, a new axis object will be created.
    plot_kws : dict
        Optional keyword arguments to include. They are sent as an argument to
        the matplotlib bar plot.
    fig_kws : dict
        Optional keyword arguments to include for the figure.

    Returns
    -------
    ax : matplotlib axis object

    Examples
    --------

    Plot the number-weighted Urban aerosol distribution

    .. plot::
        :context: close-figs

        >>> import opcsim, seaborn as sns
        >>> d = opcsim.load_distribution("Urban")
        >>> ax = opcsim.plots.cdfplot(d)
        >>> ax.set_title("Urban Aerosol Distribution", fontsize=16)
        >>> sns.despine()

    Let's plot the volume weighted version

    .. plot::
        :context: close-figs

        >>> ax = opcsim.plots.cdfplot(d, weight='volume')
        >>> ax.set_title("Volume Weighted Urban Aerosol Distribution", fontsize=16)
        >>> ax.legend(loc='best')
        >>> sns.despine()

    """
    if not hasattr(distribution, 'cdf'):
        raise Exception("Invalid AerosolDistribution.")

    if weight not in ['number', 'surface', 'volume', 'mass']:
        raise ValueError("Invalid weight: ['number', 'surface', 'volume', 'mass']")

    # Set the default dp values to plot against
    dp = kwargs.pop('dp', np.logspace(-3, 1, 1000))

    # Set the default figure kws
    default_fig_kws = dict()
    fig_kws = dict(default_fig_kws, **fig_kws)

    # Make a new axis object if one wasnt' set
    if ax is None:
        plt.figure(**fig_kws)
        ax = plt.gca()

    # Set the plot_kws as a mapping of default and kwargs
    default_plot_kws = dict(
                        alpha=1,
                        linewidth=5
                        )

    # Set the plot_kws
    plot_kws = dict(default_plot_kws, **plot_kws)

    # Plot the compete distribution
    nc = next(ax._get_lines.prop_cycler)['color']

    # If label kwarg is present, use -> otherwise use default
    label = kwargs.pop('label', distribution.label)

    data = distribution.cdf(dp, weight=weight, **kwargs)

    ax.plot(dp, data, color=nc, label=label, **plot_kws)

    ax.semilogx()
    ax.xaxis.set_major_formatter(mtick.FormatStrFormatter("%.3g"))

    ax.set_xlabel("$D_p \; [\mu m]$")
    ax.set_ylabel(YLABEL_CDF[weight])

    # Set the yaxis to show scientific notation when numbers > 1e4
    ax.yaxis.set_major_formatter(lrg_number_fmt)

    return ax


def calplot(opc, ax=None, plot_kws={}, fig_kws={}, **kwargs):
    """
    """
    # Make a new axis object if one wasnt' set
    if ax is None:
        plt.figure(**fig_kws)
        ax = plt.gca()
    
    xs = kwargs.pop("dp", np.logspace(-1.5, 1.5, 250))

    # Set the plot_kws as a mapping of default and kwargs
    default_plot_kws = dict(alpha=1, linewidth=3)

    # Set the plot_kws
    plot_kws = dict(default_plot_kws, **plot_kws)

    # compute the Cscat values
    yvals = np.array([cscat(x, wl=opc.wl, refr=opc.calibration_refr,
                            theta1=opc.theta[0], theta2=opc.theta[1]) for x in xs])
    
    nc = next(ax._get_lines.prop_cycler)['color']

    ax.plot(xs, yvals, color=nc, label="Mie", **plot_kws)
    ax.plot(opc.bin_boundaries, opc.calibration_vals, "o-",
            color=next(ax._get_lines.prop_cycler)["color"], label="Calibration")

    ax.semilogx()
    ax.semilogy()
    ax.xaxis.set_major_formatter(mtick.FormatStrFormatter("%.3g"))
    ax.set_xlabel("$D_p \; [\mu m]$")
    ax.set_ylabel("$C_{scat}\; [cm^2/particle]$")
    ax.legend(loc="upper left")

    return ax


__all__ = [
    'histplot',
    'pdfplot',
    'cdfplot',
    'calplot'
]



"""Utility functions
"""
import numpy as np
from scipy.optimize import curve_fit
import math

# assuming STP
RHO_H20 = 0.997


def make_bins(dmin, dmax, n_bins, base='log'):
    """Returns a 3xn array of bin diameters.

    Build a 3xn array of bin diameters that can then be fed directly into
    the OPC class. The default behaviour is to space the bins equally on a log10
    basis.

    Parameters
    ----------

    dmin : float
        Minimum particle diameter in microns
    dmax : float
        Maximum particle diameter in microns
    n_bins : int
        Number of bins
    base : {'log' | 'none' | None}

    Returns
    -------
    bins : array
        Returns a 3xn_bins array in the format of [left edge, midpoint, right edge]


    Examples
    --------

    Build a set of bins for an OPC with dmin=0.5, dmax=2.5, and 3 bins:

    >>> bins = opcsim.utils.make_bins(0.5, 2.5, 3)

    """
    if base is None:
        base = 'none'

    bins = np.zeros((n_bins, 2)) * np.nan

    # Set the left edge of bin0 and right edge of the last bin
    bins[0, 0] = dmin
    bins[-1, -1] = dmax

    mult = 1. / (( np.log10(dmax) - np.log10(dmin) ) / n_bins )

    for i in range(n_bins):
        bins[i, -1]  = 10**(np.log10( bins[i, 0]) + ( 1. / mult ))

        if i < n_bins - 1:
            bins[i + 1, 0] = bins[i, -1]

    return midpoints(bins, base=base)


def midpoints(bins, base='log'):
    """Returns a 3xn array of bin diameters.

    Build a 3xn array of bin diameters from a 2xn array that can then be fed
    directly into the OPC class. The default behaviour is to space the bins
    equally on a log10 basis.

    Parameters
    ----------

    bins : 2xn array
        A 2xn array with bin boundaries.
    base : {'log' | 'none' | None}

    Returns
    -------
    bins : array
        Returns a 3xn_bins array in the format of [left edge, midpoint, right edge]


    Examples
    --------

    Build a set of bins for an OPC similar to a Dylos DC1100 Pro:

    >>> arr = np.array([0.5, 2.5], [2.5, 10])
    >>> bins = opcsim.utils.midpoints(arr)

    """
    base = 'base' if base is None else base

    if bins.shape[1] != (2 or 3):
        raise ValueError("Invalid bins array. Must be either 2xn or 3xn.")

    tmp = np.zeros((bins.shape[0], 3)) * np.nan
    tmp[:, 0] = bins[:, 0]
    tmp[:, 2] = bins[:, -1]

    for i in range(tmp.shape[0]):
        if base.lower() == 'log':
            tmp[i, 1] = 10**np.log10([tmp[i,0], tmp[i, 2]]).mean()
        else:
            tmp[i, 1] = np.mean([tmp[i, 0], tmp[i, 2]])

    return tmp


def k_kohler(diam_dry, kappa=0., rh=0.):
    """Calculate the wet diameter of a particle based on the hygroscopic growth 
    parameter, kappa (k-Kohler theory).

    .. math::

        D_w=D_d*\sqrt[3]{1 + \\frac{a_w}{1-a_w}\kappa_{eff}}

    Parameters
    ----------
    diameter_dry: float
        The dry diameter in any units (nm or um most likely)
    kappa: float, optional
        The effective kappa value
    rh: float: optional
        The relative humidity as a percentage (0.0-100.0)

    Returns
    -------
    diameter_wet
        The wet diameter in the same units supplied for the dry diameter

    """
    # calculate the water activity
    aw = rh / 100.

    return diam_dry * math.pow(1 + kappa * (aw / (1 - aw)), 1./3.)


def rho_eff(rho, weights=None, diams=None):
    """Calculate the effective density of a particle by calculating the 
    wet and dry percentages and taking the weighted sum. Alternatively,
    an array of diameters can be passed which will be used to 
    calculate the volumetric weights.

    Parameters
    ----------
    rho: ndarray of floats
        An array of particle densities.
    weights: ndarray of floats
        An array of volumetric weight percentages.
    diams: ndarray of floats
        An array of diameters for each species.
    
    Returns
    -------
    rho_eff: float
        The weighted density of the wet particle.
    """
    rho = np.asarray(rho)

    # if diams are present, compute their weights
    if diams is not None:
        diams = np.asarray(diams)
        weights = (diams**3) / (diams**3).sum()

    weights = np.asarray(weights)

    return (weights * rho).sum()


def k_eff(kappas, weights=None, diams=None):
    """Calculate the effective k-kohler coefficient from 
    an array of kappa values and their weights. Alternatively,
    an array of diameters can be passed which will be used to 
    calculate the volumetric weights.

    .. math::

        \kappa=\sum_{i=1} \epsilon_i \kappa_i

    Parameters
    ----------
    kappas: ndarray
        An array of k-kohler coefficients.
    weights: ndarray, optional
        An array of volumetric weights.
    diams: ndarray
        An array of diameters used to calculate the weights.
    
    Returns
    -------
    k_eff: float
        The effective k-kohler coefficient.
    """
    kappas = np.asarray(kappas)

    # if diams are present, compute their weights
    if diams is not None:
        diams = np.asarray(diams)
        weights = (diams**3) / (diams**3).sum()

    weights = np.asarray(weights)

    return (kappas * weights).sum()


def ri_eff(species, weights=None):
    """Calculate the effective refractive index for an 
    array of refractive indices and their respective weights. 
    Alternatively, an array of diameters can be passed which 
    will be used to calculate the volumetric weights.

    .. math::

        n_{eff}=\sum_{i=1}^{N} \\frac{V_i}{V_{total}}*n_i

    Parameters
    ----------
    species: ndarray
        An array of refractive indices.
    weights: ndarray, optional
        An array of volumetric weights.
    
    Returns
    -------
    ri_eff: float
        The effective refractive index.
    """
    species = np.asarray(species)
    weights = np.asarray(weights)

    # calculate the real and imag parts separately
    real = (species.real * weights).sum()
    imag = (species.imag * weights).sum()

    return complex(real, imag)


def power_law_fit(diams, cscat, fit_kws={}):
    """Generate a power-law fit (linear in log-log space)
    between bin midpoints and the scattering cross-section.

    Parameters
    ----------
    diams: ndarray
        An array of diameters corresponding to the boundaries of each bin.
    cscat: ndarray
        An array of Cscat values corresponding to the boundaries of each bin.
    fit_kws: dict
        A dictionary of keyword arguments that is passed directly to scipy.curve_fit
        for the optimization. Please see the scipy.optimize.curve_fit docs for details.

    Returns
    -------
    rv: ndarray
        An array of fitted Cscat values for each bin boundary diameter.

    Examples
    --------

    """
    def f(dp, a, b):
        return a*np.power(dp, b)
    
    # fit the function
    popt, _ = curve_fit(f, diams, cscat, **fit_kws)

    return f(diams, *popt)


def squash_dips(cscat_vals):
    """Remove any dips in an array by interpolating 
    around those points. If there are no dips and all 
    points are monotonically increasing, the same 
    array should be returned.

    Parameters
    ----------
    cscat: ndarray
        An array of scattering cross-section values.
    
    Returns
    -------
    rv: ndarray
        An array of smoothed scattering cross-section values.

    """
    cpy = cscat_vals.copy()

    if (np.diff(cpy) < 0).any():
        idx = np.where(np.diff(cpy) < 0)[0]

        if idx.shape[0] > 0:
            for i in np.nditer(idx):
                cpy[i] = np.mean([cpy[i-1], cpy[i+1]])
    
    return cpy



"""Make changes to the default styling which seaborn doesn't support
"""
import matplotlib as mpl

def set(**kwargs):
    mpl.rcParams['figure.autolayout'] = kwargs.pop('figure.autolayout', False)
    mpl.rcParams['mathtext.default'] = kwargs.pop('mathtext.default', 'regular')

    return



"""Contains the scoring algorithms used in the model.
"""

import numpy as np
import pandas as pd
from .models import OPC
from .utils import k_kohler, ri_eff
from .mie import cscat


def compute_bin_assessment(opc, refr, kappa, rh_values=[0., 35., 95.]):
    """Assess the ability of an OPC to assign particles to their correct bin.

    Parameters
    ----------
    opc: opcsim.OPC
    refr: complex
        The complex refractive index of the material to assess
    kappa: float
        The kappa value to use for hygroscopic growth
    rh_values: list-like
        A list of relative humidities to assess the OPC at.

    Returns
    -------
    rv: pd.DataFrame
        A dataframe containing the results with self-explanatory columns.

    Examples
    --------

    """
    assert(isinstance(opc, OPC)), "opc must be an instance of the opcsim.OPC class"

    # init the dataframe to hold our results
    rv = pd.DataFrame()

    for rh in rh_values:
        for i, _bins in enumerate(opc.bins):
            # compute the wet diameter
            wet_diam_lo = k_kohler(diam_dry=_bins[0], kappa=kappa, rh=rh)
            wet_diam_hi = k_kohler(diam_dry=_bins[-1], kappa=kappa, rh=rh)

            # compute the pct_dry
            pct_dry = (_bins[0]**3) / (wet_diam_lo**3)

            # compute the effective RI
            ri = ri_eff(species=[refr, complex(1.333, 0)], weights=[pct_dry, 1-pct_dry])

            # compute the scattering cross-section
            cscat_lo_exp = cscat(
                dp=_bins[0], wl=opc.wl, refr=refr, theta1=opc.theta[0], theta2=opc.theta[1])
            cscat_hi_exp = cscat(
                dp=_bins[-1], wl=opc.wl, refr=refr, theta1=opc.theta[0], theta2=opc.theta[1])

            cscat_lo = cscat(
                dp=wet_diam_lo, wl=opc.wl, refr=ri, theta1=opc.theta[0], theta2=opc.theta[1])
            cscat_hi = cscat(
                dp=wet_diam_hi, wl=opc.wl, refr=ri, theta1=opc.theta[0], theta2=opc.theta[1])

            # assign bins
            bin_assign_lo = opc.calibration_function(values=[cscat_lo])
            bin_assign_hi = opc.calibration_function(values=[cscat_hi])

            # add results to the dataframe
            rv = rv.append({
                "bin_true": i,
                "bin_lo": bin_assign_lo[0] if len(bin_assign_lo) > 0 else -99,
                "bin_hi": bin_assign_hi[0] if len(bin_assign_hi) > 0 else -99,
                "refr_eff": ri,
                "rh": rh,
                "cscat_hi_ratio": cscat_hi / cscat_hi_exp,
                "cscat_lo_ratio": cscat_lo / cscat_lo_exp,
            }, ignore_index=True)
    
    # force datatypes to be correct
    rv["bin_true"] = rv["bin_true"].astype(int)
    rv["bin_lo"] = rv["bin_lo"].astype(int)
    rv["bin_hi"] = rv["bin_hi"].astype(int)
    rv["rh"] = rv["rh"].astype(float)
    rv["cscat_hi_ratio"] = rv["cscat_hi_ratio"].astype(float)
    rv["cscat_lo_ratio"] = rv["cscat_lo_ratio"].astype(float)

    return rv



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
This file contains all of the probability distribution functions
"""
import math
import numpy as np

def dn_ddp(dp, n, gm, gsd):
    """Evaluate the number distribution as a lognormal PDF.

    The PDF of a lognormal distribution is calculated using equation 8.34
    from Seinfeld and Pandis.

    .. math::

        n_N(D_p)=\\frac{dN}{dD_p}=\\frac{N_t}{\sqrt{2π}D_p lnσ_g}exp\Big(-\\frac{(lnD_p - lnD̄_{pg})^2}{2ln^2σ_g}\Big)

    Parameters
    ----------
    dp : float or an array of floats
        Particle diameter in microns.
    n : float
        Total aerosol number concentration in units of #/cc
    gm : float
        Median particle diameter (geometric mean) in units of microns.
    gsd : float
        Geometric Standard Deviation of the distribution.

    Returns
    -------
    n(Dp) | float or an array of floats
        Returns the total number of particles at diameter dp in units of
        :math:`\mu m^{-1} cm^{-3}`.

    See Also
    --------
    opcsim.equations.pdf.dn_dlndp
    opcsim.equations.pdf.dn_dlogdp
    opcsim.equations.cdf.nt

    """
    res = (n / (np.sqrt(2*np.pi)*dp*np.log(gsd))) * \
            np.exp(-(np.log(dp) - np.log(gm))** 2 / (2*np.log(gsd)** 2))

    return res

def dn_dlndp(dp, n, gm, gsd):
    """The PDF of a lognormal distribution as calculated using equation 8.33 by way
    of 8.21 from Seinfeld and Pandis.

    .. math::

        n_N^e(lnD_p)=\\frac{dN}{dlnD_p}=\\frac{N_t}{\sqrt{2π} lnσ_g}exp(-\\frac{(lnD_p - lnD̄_{pg})^2}{2ln^2σ_g})

    Parameters
    ----------
    dp : float or an array of floats
        Particle diameter in microns.
    n : float
        Total aerosol number concentration in units of #/cc
    gm : float
        Median particle diameter (geometric mean) in units of microns.
    gsd : float
        Geometric Standard Deviation of the distribution.

    Returns
    -------
    dn/dlndp | float
        Returns the total number of particles at diameter dp in units of
        [cm-3]

    See Also
    --------
    opcsim.equations.pdf.dn_ddp
    opcsim.equations.pdf.dn_dlogdp
    opcsim.equations.cdf.nt

    """
    return dp * dn_ddp(dp, n, gm, gsd)

def dn_dlogdp(dp, n, gm, gsd):
    """The PDF of a lognormal distribution on a log10 basis as calculated using
    equation 8.18 from Seinfeld and Pandis.

    .. math::

        n_N^o(logD_p)=\\frac{dN}{dlogD_p}=ln(10)D_pn_N(D_p)

    Parameters
    ----------
    dp : float or array of floats
        Particle diameter in microns.
    n : float
        Total aerosol number concentration in units of #/cc
    gm : float
        Median particle diameter (geometric mean) in units of microns.
    gsd : float
        Geometric Standard Deviation of the distribution.

    Returns
    -------
    dn/dlogdp | float
        Returns the total number of particles at diameter dp in units of
        [cm-3]

    See Also
    --------
    opcsim.equations.pdf.dn_ddp
    opcsim.equations.pdf.dn_dlndp
    opcsim.equations.cdf.nt

    """
    return np.log(10) * dp * dn_ddp(dp, n, gm, gsd)

def ds_ddp(dp, n, gm, gsd):
    """The surface-area weighted PDF of a lognormal distribution as calculated
    using equation 8.4 from Seinfeld and Pandis.

    .. math::

        n_s(D_p)=πD_p^2n_N(D_p)

    Parameters
    ----------
    dp : float or array of floats
        Particle diameter in microns.
    n : float
        Total aerosol number concentration in units of #/cc
    gm : float
        Median particle diameter (geometric mean) in units of microns.
    gsd : float
        Geometric Standard Deviation of the distribution.

    Returns
    -------
    ds/ddp | float
        Returns the surface area of particles at diameter dp in units of
        [um cm-3]

    See Also
    --------
    opcsim.equations.pdf.ds_dlndp
    opcsim.equations.pdf.ds_dlogdp
    opcsim.equations.cdf.st

    """
    return np.pi * dp ** 2 * dn_ddp(dp, n, gm, gsd)

def ds_dlndp(dp, n, gm, gsd):
    """The surface-area weighted PDF of a lognormal distribution as calculated
    using equation 8.10 from Seinfeld and Pandis.

    .. math::

        n_s^e(ln D_p)=πD_p^2n_N^e(D_p)

    Parameters
    ----------
    dp : float or array of floats
        Particle diameter in microns.
    n : float
        Total aerosol number concentration in units of #/cc
    gm : float
        Median particle diameter (geometric mean) in units of microns.
    gsd : float
        Geometric Standard Deviation of the distribution.

    Returns
    -------
    ds/dlndp | float
        Returns the surface area of particles at diameter dp on a log basis
        in units of [um2 cm-3]

    See Also
    --------
    opcsim.equations.pdf.ds_ddp
    opcsim.equations.pdf.ds_dlogdp
    opcsim.equations.cdf.st

    """
    return np.pi * dp ** 2 * dn_dlndp(dp, n, gm, gsd)

def ds_dlogdp(dp, n, gm, gsd):
    """The surface-area weighted PDF of a lognormal distribution as calculated
    using equation 8.19 from Seinfeld and Pandis.

    .. math::

        n_s^o(log D_p)=ln(10)D_p n_s(D_p)

    Parameters
    ----------
    dp : float or array of floats
        Particle diameter in microns.
    n : float
        Total aerosol number concentration in units of #/cc
    gm : float
        Median particle diameter (geometric mean) in units of microns.
    gsd : float
        Geometric Standard Deviation of the distribution.

    Returns
    -------
    ds/dlogdp | float
        Returns the surface area of particles at diameter dp on a log10 basis
        in units of [um2 cm-3]

    See Also
    --------
    opcsim.equations.pdf.ds_ddp
    opcsim.equations.pdf.ds_dlndp
    opcsim.equations.cdf.st

    """
    return np.log(10) * dp * ds_ddp(dp, n, gm, gsd)

def dv_ddp(dp, n, gm, gsd):
    """The volume weighted PDF of a lognormal distribution as calculated
    using equation 8.6 from Seinfeld and Pandis.

    .. math::

        n_V(D_p)=\\frac{π}{6}D_p^3 n_N(D_p)

    Parameters
    ----------
    dp : float or array of floats
        Particle diameter in microns.
    n : float
        Total aerosol number concentration in units of #/cc
    gm : float
        Median particle diameter (geometric mean) in units of microns.
    gsd : float
        Geometric Standard Deviation of the distribution.

    Returns
    -------
    dv/ddp | float
        Returns the volume of particles at diameter dp
        in units of [um2 cm-3]

    See Also
    --------
    opcsim.equations.pdf.dv_dlndp
    opcsim.equations.pdf.dv_dlogdp
    opcsim.equations.cdf.vt

    """
    return (np.pi / 6.) * dp ** 3 * dn_ddp(dp, n, gm, gsd)

def dv_dlndp(dp, n, gm, gsd):
    """The volume weighted PDF of a lognormal distribution as calculated
    using equation 8.10 from Seinfeld and Pandis.

    .. math::

        n_V^e(ln D_p)=\\frac{π}{6}D_p^3 n_N^e(ln D_p)

    Parameters
    ----------
    dp : float or array of floats
        Particle diameter in microns.
    n : float
        Total aerosol number concentration in units of #/cc
    gm : float
        Median particle diameter (geometric mean) in units of microns.
    gsd : float
        Geometric Standard Deviation of the distribution.

    Returns
    -------
    dV/dlndp | float
        Returns the volume of particles at diameter dp on a log basis
        in units of [um3 cm-3]

    See Also
    --------
    opcsim.equations.pdf.dv_ddp
    opcsim.equations.pdf.dv_dlogdp
    opcsim.equations.cdf.vt

    """
    return (np.pi / 6.) * dp ** 3 * dn_dlndp(dp, n, gm, gsd)

def dv_dlogdp(dp, n, gm, gsd):
    """The volume weighted PDF of a lognormal distribution as calculated
    using equation 8.20 from Seinfeld and Pandis.

    .. math::

        n_V^o(log D_p)=log(10)*D_p n_V(D_p)

    Parameters
    ----------
    dp : float or array of floats
        Particle diameter in microns.
    n : float
        Total aerosol number concentration in units of #/cc
    gm : float
        Median particle diameter (geometric mean) in units of microns.
    gsd : float
        Geometric Standard Deviation of the distribution.

    Returns
    -------
    dV/dlogdp | float
        Returns the volume of particles at diameter dp on a log10 basis
        in units of [um3 cm-3]

    See Also
    --------
    opcsim.equations.pdf.dv_ddp
    opcsim.equations.pdf.dv_dlndp
    opcsim.equations.cdf.vt

    """
    return np.log(10) * dp * dv_ddp(dp, n, gm, gsd)



import numpy as np
import pandas as pd
import math
import scipy

from .distributions import AerosolDistribution
from .utils import make_bins, midpoints, squash_dips, power_law_fit, \
    ri_eff, rho_eff, k_kohler
from .mie import cscat
import functools

RI_COMMON = {
    "psl": complex(1.59, 0),
    "ammonium_sulfate": complex(1.521, 0),
    "sodium_chloride": complex(1.5405, 0),
    "sodium_nitrate": complex(1.448, 0),
    "black_carbon": complex(1.95, 0.79),
    "sulfuric_acid": complex(1.427, 0),
    "soa": complex(1.4, 0.002),
    "h2o": complex(1.333, 0.),
    "urban_low": complex(1.6, 0.034),
    "urban_high": complex(1.73, 0.086)
}


class OPC(object):
    """Define an Optical Particle Counter (OPC) with unique properties 
    for wavelength, bins, and viewing angle.
    """
    def __init__(self, wl, bins=None, n_bins=5, dmin=0.5, 
            dmax=2.5, theta=(30., 90.), **kwargs):
        """

        Parameters
        ----------
        wl: float
            The laser wavelength in units of microns.
        bins: 3xn array
            An array of bin diameters containing the (left boundary, midpoint, right boundary).
        n_bins: int
            The number of desired bins. This should be used with dmin and dmax to generate 
            a 3xn array of bins.
        dmin: float
            The left-most bin boundary of the OPC.
        dmax: float
            The right-most bin boundary of the OPC.
        theta: tuple of floats
            The viewing range in units of degrees.
        
        Returns
        -------
        OPC:
            An instance of the OPC class.
        
        Examples
        --------

        Initialize an OPC with 5 bins between 0.5 - 2.5 microns.

        >>> opc = opcsim.OPC(wl=0.658, n_bins=5, dmin=0.5, dmax=2.5, theta=(30., 90.))

        Initialize an OPC with known bins as defined by its bin boundaries.
        
        >>> opc = opcsim.OPC(wl=0.658, bins=[0.38, 0.54, 0.78, 1.05, 1.5, 2.5], theta=(32., 88.))

        """
        # set some params
        self.n_bins = n_bins
        self.dmin = dmin
        self.dmax = dmax
        self.bins = bins
        self.wl = wl
        self.theta = theta
        self.label = kwargs.pop("label", None)
        self.calibration_function = None
        self.calibration_refr = None
        self.calibration_vals = None
        self._cscat_boundaries = None

        # generate the bins
        if self.bins is None:
            self.bins = make_bins(self.dmin, self.dmax, self.n_bins)
        else:
            self.bins = np.asarray(self.bins)
            
            if self.bins.ndim == 1:
                self.bins = midpoints(np.array([self.bins[0:-1], self.bins[1:]]).T)
            
            # reset a few fields
            self.n_bins = self.bins.shape[0]
            self.dmin = self.bins[0, 0]
            self.dmax = self.bins[-1, -1]
        
        self.bin_boundaries = np.append(self.bins[:, 0], self.bins[-1, -1])

        return

    @property
    def dlogdp(self):
        """Log-weighted bin width in microns"""
        return np.log10(self.bins[:, -1]) - np.log10(self.bins[:, 0])

    @property
    def ddp(self):
        """Bin width in microns"""
        return self.bins[:, -1] - self.bins[:, 0]

    @property
    def midpoints(self):
        """Bin midpoints in microns"""
        return self.bins[:, 1]

    def calibrate(self, material, method='spline', mie_kws={}, fit_kws={}):
        """Calibrate the OPC assuming a specific material. 

        By calibration, we mean a method used to relate the peak height, 
        which is related to the scattering cross-section (Cscat) to the 
        particle size. At its simplest, we use this function to determine 
        which bin a particle belongs to based on its Cscat value. Once 
        calibrated, a calibration function is saved as a digitizer which 
        will take as an input a Cscat value and return the bin number of 
        the OPC it belongs to.

        Parameters
        ----------
        material: string or complex number
            Either a string containing the material available in the
            lookup table, or the complex refractive index. The option for 
            lookup values are: ['ammonium_sulfate', 'bc', 'psl']. Since 
            the refractive index is wavelength dependant, it is 
            recommended you define the refractive index at your OPC's 
            wavelength if you are looking for the best results.
        method: string or callable
            The method to use for creating a calibration curve. Options 
            include (1) 'spline' which removes any non-monotonicly increasing 
            points from the Cscat to Dp curve; or (2) 'linear' fits a linear model 
            (in log-log space) between Cscat and Dp.
        mie_kws: dict
            Optional dictionary containing keyword arguments that is sent 
            directly to opcsim.mie.cscat when computing the scattering 
            cross section values used in the optimization.
        fit_kws: dict
            Optional dictionary containing keyword arguments that is 
            sent directly to scipy.optimize.curve_fit when generating 
            a fit using the 'fit_linear' approach. Please see the 
            scipy.optimize.curve_fit docs for more details.

        Examples
        --------

        Calibrate an OPC using PSL's

        >>> opc = opcsim.OPC(n_bins=5)
        >>> opc.calibrate(material="psl", method="spline")

        Calibrate an OPC using a custom material
        
        >>> opc = opcsim.OPC(n_bins=5)
        >>> opc.calibrate(material=complex(1.9, 0.4), method="spline")

        Calibrate an OPC where the calibration curve is a fitted line (PSL's)
        
        >>> opc = opcsim.OPC(n_bins=5)
        >>> opc.calibrate(material="psl", method="linear")

        """
        # determine the complex refractive index
        if type(material) == str:
            try:
                refr = RI_COMMON[material.lower()]
            except:
                raise ValueError("This material does not exist in our (limited) database.")
        else:
            if type(material) != complex:
                refr = complex(material, 0)
            else:
                refr = material
        
        # calculate Cscat at all bin boundaries
        yvals = np.array([cscat(dp=dp, wl=self.wl, refr=refr, theta1=self.theta[0],
                                theta2=self.theta[1], **mie_kws) for dp in self.bin_boundaries])
        
        # generate the fitted Cscat values based on the method chosen
        if method == "spline":
            yvals = squash_dips(yvals)
        elif method == "linear":
            # define the function we fit dp to Cscat
            def f(dp, a, b):
                return a * np.power(dp, b)
            
            # if log-weight, set sigma to weight in log-units
            fit_kws["sigma"] = np.power(10, np.log10(yvals) + 1)

            # fit the data with optional kwargs
            popt, _ = scipy.optimize.curve_fit(f, self.bin_boundaries, yvals, **fit_kws)

            # set the yvals to the fitted values
            yvals = f(self.bin_boundaries, *popt)
        elif method == "piecewise":
            # define the function to fit
            def f(x, x0, a1, a2, b1, b2):
                return np.piecewise(x, [x < x0], [
                    lambda x: a1*np.power(x, b1),
                    lambda x: a2*np.power(x, b2)
                ])
            
            # if log-weight, set sigma
            fit_kws["sigma"] = np.power(10, np.log10(yvals) + 1)

            # fit the data with optional kwargs
            popt, _ = scipy.optimize.curve_fit(f, self.bin_boundaries, yvals, **fit_kws)

            # set the yvals to the fitted vals
            yvals = f(self.bin_boundaries, *popt)
        else:
            raise ValueError("The `method` chosen is not currently supported.")

        # generate the digitizer and set as the calibration function
        # the calibration function must take just one argument (cscat(s)) and return 
        # a bin number(s)
        self.calibration_function = functools.partial(self._digitize_opc_bins, cscat_boundaries=yvals)
        self.calibration_refr = refr
        self.calibration_vals = yvals

        # save the fitted boundaries for potential future use
        self._cscat_boundaries = yvals

    def evaluate(self, distribution, rh=0., **kwargs):
        """Return the total number of particles in each bin for a given AerosolDistribution.

        We evaluate an OPC for a given distribution by calculating the Cscat value
        for every particle in the distribution and assigning it to a bin of the OPC.
        Since we are beginning with a PDF and not a distribution, we must first 
        discretize our PDF - we can choose the number of bins to convert of PDF 
        into which allows us to limit the computation needed by only performing 
        calculations for a small subset of the actual particles (i.e. we can do one
        calculation for a tiny bin and then replicate it without needing to re-do the 
        Mie calculations).

        Parameters
        ----------
        distribution: AerosolDistribution
            A valid instance of the AerosolDistribution class.
        rh: float
            The relative humidity in % (0-100)

        Returns
        -------
        dN: array
            The number of particles in each OPC bin (size is the number of bins)

        Examples
        --------

        Evaluate an OPC for the Urban distribution

        >>> opc = opcsim.OPC(wl=0.658, n_bins=5)
        >>> opc.calibrate(material="psl")
        >>> d = opcsim.load_distribution("urban")
        >>> vals = opc.evaluate(d, rh=0.)

        Evaluate a distribution of Ammonium Sulfate at various RH's
        
        >>> opc = opcsim.OPC(wl=0.658, n_bins=5)
        >>> d = opcsim.AerosolDistribution()
        >>> d.add_mode(n=1000, gm=500e-3, gsd=1.5, kappa=0.53, refr=complex(1.521, 0), rho=1.77)
        >>> vals_0 = opc.evaluate(d, rh=0.)
        >>> vals_50 = opc.evaluate(d, rh=50.)
        >>> als_100 = opc.evaluate(d, rh=100.)

        """
        if not self.calibration_function:
            raise Exception("The OPC must be calibrated before computing a histogram.")

        # create the bins for the distribution PDF
        bounds = kwargs.pop("bounds", (self.dmin / 2, 10.))

        # calculate the total number of particles in the distribution
        ntot = distribution.cdf(dmin=0, dmax=100., weight='number')

        bounds = np.logspace(start=np.log10(bounds[0]), stop=np.log10(
            bounds[1]), num=kwargs.pop("n_bins", int(min(ntot/3, 250))))

        # we need to calculate the midpoint of the new bounds to compute the cscats!
        diams = np.mean([bounds[:-1], bounds[1:]], axis=0)

        # for each mode...
        rv = np.zeros(self.n_bins)

        for m in distribution.modes:
            # divide our PDF into bins to make the computations a bit easier
            n = np.array([distribution.cdf(dmin=a, dmax=b, mode=m["label"], rh=rh)
                                for a, b in zip(bounds[0:-1], bounds[1:])])
            
            # calculate the % dry based on hygroscopic growth
            pct_dry = 1. / (k_kohler(diam_dry=1., kappa=m["kappa"], rh=rh)**3)

            refr = ri_eff([m["refr"], RI_COMMON['h2o']], weights=[pct_dry, 1-pct_dry])

            # iterate over each bin and calculate the Cscat value and build an array
            for dp, dn in zip(diams, n):
                # ammend the RI based on the RH
                v = cscat(
                    dp, wl=self.wl, refr=refr, theta1=self.theta[0], theta2=self.theta[1])
                
                bin_assign = self.calibration_function(values=[v])

                if len(bin_assign) > 0:
                    rv[bin_assign] += dn
        
        return rv
    
    def histogram(self, distribution, weight="number", base="log10", rh=0., **kwargs):
        """Return a histogram containing the [weight] of particles in each OPC bin.

        This represents what the OPC 'sees'. All calculations are made assuming the 
        center of the bin is reprentative of the entire bin.

        Parameters
        ----------
        distribution: AerosolDistribution
        weight : {'number' | 'surface' | 'volume'}
            Choose how to weight the pdf. Default is `number`.
        base : {'none' | 'log10'}
            Base algorithm to use. Default is 'log10'.
        rh: float
            The relative humidity in percent (0-100).

        Returns
        -------
        d[weight]/d[base]Dp : array
            Returns an array with the evaluated
            PDF. This data can be
            directly plotted as a histogram using matplotlib bar plots. By 
            default, dN/dlogDp is returned.

        Examples
        --------

        Evaluate an OPC for the Urban distribution and return dN/dlogDp

        >>> opc = opcsim.OPC(n_bins=5)
        >>> opc.calibrate(material="psl")
        >>> d = opcsim.load_distribution("urban")
        >>> hist = opc.histogram(d, weight="number", rh=0.)

        Evaluate an OPC for the Urban distribution and return dV/dlogDp

        >>> opc = opcsim.OPC(n_bins=5)
        >>> opc.calibrate(material="psl")
        >>> d = opcsim.load_distribution("urban")
        >>> hist = opc.histogram(d, weight="volume", rh=0.)

        Evaluate an OPC for the Urban distribution and return dN/dDp

        >>> opc = opcsim.OPC(n_bins=5)
        >>> opc.calibrate(material="psl")
        >>> d = opcsim.load_distribution("urban")
        >>> hist = opc.histogram(d, weight="number", base=None, rh=0.)

        Evaluate a distribution of Ammonium Sulfate at 50% RH
        
        >>> opc = opcsim.OPC(n_bins=5)
        >>> d = opcsim.AerosolDistribution()
        >>> d.add_mode(n=1000, gm=500e-3, gsd=1.5, kappa=0.53, refr=complex(1.521, 0), rho=1.77)
        >>> hist = opc.histogram(d, weight="number", base="log10", rh=50.)

        """
        # get the density if needed [units of g/cc]
        rho = kwargs.pop("rho", 1.65)

        # get the binned values in units of dN/bin
        rv = self.evaluate(distribution, rh=rh, **kwargs)

        if weight not in ["number", "surface", "volume", "mass"]:
            raise ValueError("Invalid `weight` parameter")

        # convert the values into the proper weight
        if weight == "surface":
            rv *= (np.pi * self.midpoints**2)
        elif weight == "volume":
            rv *= (self.midpoints**3*np.pi/6.)  # dp3 * pi/6*N
        elif weight == "mass":
            rv *= rho * (self.midpoints**3*np.pi/6.)  # rho*dp3 * pi/6*N
        else:
            pass

        # divide the values by the proper base
        if base == "log10":
            rv = rv / self.dlogdp
        else:
            rv = rv / self.ddp

        return rv
    
    def integrate(self, distribution, dmin=0., dmax=1., weight="number", rh=0., **kwargs):
        """Integrate the distribution according to the OPC for any [weight].

        By default, this method returns the total number of particles between 0-1 microns.
        It can be used to calculate PM values at any breakpoint. When calculating a PM value
        for a breakpoint that falls within a bin, a simple percentage of the bin is used. For
        example, if you calculate PM1 and the bin that contains 1-micron actually covers 0.5 
        to 1.5 microns, then 1/2 the mass from that bin will count towards the PM1 value.

        Parameters
        ----------
        distribution: AerosolDistribution
            The aerosol distribution to be evaluated.
        dmin: float
            The minimum particle diameter [microns] to integrate from.
        dmax: float
            The maximum particle diameter [microns] to integrate to.
        weight : {'number' | 'surface' | 'volume'}
            Choose how to weight the pdf. Default is `number`.
        rh: float
            The relative humidity in percent (0-100).

        Returns
        -------
        rv: float
            The total [weight] between dmin and dmax. By default, 
            the total number of particles (i.e. weight='number') 
            are returned.

        Examples
        --------

        Calculate the total number of particles between 0-1 microns
        for the Urban distribution.

        >>> opc = opcsim.OPC(n_bins=5)
        >>> opc.calibrate(material="psl")
        >>> d = opcsim.load_distribution("urban")
        >>> ntot = opc.integrate(d, dmin=0., dmax=1., weight="number", rh=0.)

        Calculate PM1 for the Urban Distribution when RH = 0%

        >>> opc = opcsim.OPC(n_bins=5)
        >>> opc.calibrate(material="psl")
        >>> d = opcsim.load_distribution("urban")
        >>> ntot = opc.integrate(d, dmin=0., dmax=1., weight="mass", rh=0., rho=1.5)

        Calculate PM2.5 for the Urban Distribution when RH = 50%

        >>> opc = opcsim.OPC(n_bins=5)
        >>> opc.calibrate(material="psl")
        >>> d = opcsim.load_distribution("urban")
        >>> ntot = opc.integrate(d, dmin=0., dmax=2.5, weight="mass", rh=50., rho=1.5)

        """
        rho = kwargs.pop("rho", 1.65)

        # calculate dN
        rv = self.evaluate(distribution, rh=rh, **kwargs)

        if weight not in ["number", "surface", "volume", "mass"]:
            raise ValueError("Invalid argument for `weight`")
        
        if weight == "surface":
            # convert dN to dS
            rv *= (np.pi * self.midpoints**2)
        elif weight == "volume":
            # convert dN to dV
            rv *= (self.midpoints**3*np.pi/6.)
        elif weight == "mass":
            # convert dN to dM
            rv *= rho*(self.midpoints**3*np.pi/6.)
        else:
            pass

        # cleave the array
        # for bins that are outside diameters of interest, set to 0
        # for bins that are partially within the range of diams, set to 
        # a fraction based on % within range
        factors = np.zeros(self.bins.shape[0])
        for i, b in enumerate(self.bins):
            if (b[0] >= dmin) and (b[-1] <= dmax):
                factors[i] = 1.
            elif (dmin >= b[0]) and (dmax <= b[-1]):
                factors[i] = (dmax - dmin) / (b[-1] - b[0])
            elif (dmin >= b[0]) and (dmax >= b[-1]):
                factors[i] = (b[-1] - dmin) / (b[-1] - b[0])
            elif (dmax >= b[0]) and (dmax <= b[-1]):
                factors[i] = (dmax - b[-1]) / (b[-1] / b[0])
            else:
                factors[i] = 0.

        # return the sum
        return (rv*factors).sum()

    def _digitize_opc_bins(self, cscat_boundaries, values):
        """Return the bin (or bins) corresponding to the :math:`C_{scat}` value(s).

        Parameters
        ----------
        cscat_boundaries: array-like
            An array of Cscat values at every bin boundary. The length of 
            the array should be one greater than the number of bins.
        values: array-like
            Either a single-value or array of computed Cscat values 
            that need to be assigned to OPC bins.

        Returns
        -------
        bins: np.ndarray
            An array of assigned bins associated with every value. If 
            a value was too small (i.e. value < smallest in cscat_boundaries)
            or too large, it will be dropped. Thus, the size of bins is less than 
            or equal to the size of values.

        """
        _binb = np.asarray(cscat_boundaries)
        _vals = np.asarray(values)

        # create a digitized version of all values. 0 implies it is too small and
        # values that are too large will be assigned the value for the length of the array
        digitized = np.digitize(_vals, bins=_binb)

        # keep only valid bins
        digitized = digitized[np.where(
            (digitized != 0) & (digitized != len(_binb)))]

        # subtract 1 from every bin so they are 0-indexed
        digitized -= 1

        return digitized

    def __repr__(self): # pragma: no cover
        return str(self.__class__)


class Nephelometer(object):
    """Define a Nephelometer by its wavelength and range of viewing angles.
    """
    def __init__(self, wl, theta=(7., 173.), **kwargs):
        """
        Parameters
        ----------
        wl: float
            The wavelength of laser used in the device
        theta: tuple of floats
            The viewing angle range in degrees

        Returns
        -------
        Nephelometer
            An instance of the opcsim.Nephelometer class

        Examples
        --------

        Build a simple nephelometer

        >>> neph = opcsim.Nephelometer(wl=0.658)

        """
        self.wl = wl
        self.theta = theta
        self.pm1_ratio = None
        self.pm25_ratio = None
        self.pm10_ratio = None
    
    def calibrate(self, distribution, rh=0., **kwargs):
        """Calibrate a Nephelometer using any AerosolDistribution.

        Parameters
        ----------
        distribution: opcsim.AerosolDistribution
            The aerosol distribution used to calibrate the Nephelometer.
        rh: float
            The relative humidity at which the calibration takes place.
            Default is 0 %.

        Returns
        -------
        None

        Examples
        --------

        Build and calibrate a Nephelometer to a simple 1-mode distribution of Ammonium Sulfate.

        >>> neph = Nephelometer(wl=0.6, theta=(7., 173.))
        >>>
        >>> d = opcsim.AerosolDistribution("AmmSulf")
        >>> d.add_mode(n=1000, gm=0.2, gsd=1.25, refr=complex(1.521, 0), rho=1.77, kappa=0.53)
        >>> 
        >>> neph.calibrate(d, rh=0.)

        """
        n_bins = kwargs.pop("n_bins", 100)

        # compute the PM1, PM25, and PM10 masses
        pm1 = distribution.cdf(dmin=0., dmax=1., weight="mass")
        pm25 = distribution.cdf(dmin=0., dmax=2.5, weight="mass")
        pm10 = distribution.cdf(dmin=0., dmax=10., weight="mass")

        # compute the total scattered light across the distribution
        total_cscat = self._sum_across_distribution(distribution, n_bins=n_bins, rh=rh)

        # set the ratios
        self.pm1_ratio = total_cscat / pm1
        self.pm25_ratio = total_cscat / pm25
        self.pm10_ratio = total_cscat / pm10

    def _sum_across_distribution(self, distribution, n_bins=100, rh=0., **kwargs):
        """
        """
        total_cscat = 0.

        for m in distribution.modes:
            gm = m["GM"]
            gsd = m["GSD"]
            refr = m["refr"]

            # alter the GM and RI per the RH specifed
            gm = k_kohler(diam_dry=gm, kappa=m["kappa"], rh=rh)

            pct_dry = m["GM"]**3 / gm**3

            refr = ri_eff(species=[refr, RI_COMMON["h2o"]], weights=[pct_dry, 1-pct_dry])

            # compute the bounds
            bounds = np.logspace(start=np.log10(gm/(gsd**4)), stop=np.log10(gm*(gsd**4)), num=n_bins)

            # compute the midpoints of each bin
            midpoints = np.mean([bounds[:-1], bounds[1:]], axis=0)

            # compute the number of particles in each bin
            n = np.array([distribution.cdf(dmin=a, dmax=b, mode=m["label"], rh=rh)
                          for a, b in zip(bounds[:-1], bounds[1:])])
            
            # compute the mean Cscat for each bin
            mean_cscat = np.array([cscat(dp, wl=self.wl, refr=refr,
                                    theta1=self.theta[0], theta2=self.theta[1]) for dp in midpoints])
            
            # add to the running total
            total_cscat += (n*mean_cscat).sum()

        return total_cscat
    
    def evaluate(self, distribution, rh=0., **kwargs):
        """Evaluate a Nephelometer for an AerosolDistribution at 
        a given relative humidity.

        Parameters
        ----------
        distribution: opcsim.AerosolDistribution
            The aerosol distribution used to calibrate the Nephelometer.
        rh: float
            The relative humidity at which the calibration takes place.
            Default is 0 %.

        Returns
        -------
        cscat: float
            The total summed scattering signal across the entire size distribution.
        pm1: float
            The total PM1 value in ug/m3
        pm25: float
            The total PM2.5 value in ug/m3
        pm10: float
            The total PM10 value in ug/m3
        Cscat, pm1, pm25, pm10: floats
            

        Examples
        --------

        Build and calibrate a Nephelometer to a simple 1-mode distribution of Ammonium Sulfate.

        >>> neph = Nephelometer(wl=0.6, theta=(7., 173.))
        >>>
        >>> d = opcsim.AerosolDistribution("AmmSulf")
        >>> d.add_mode(n=1000, gm=0.2, gsd=1.25, refr=complex(1.521, 0), rho=1.77, kappa=0.53)
        >>> 
        >>> neph.calibrate(d, rh=0.)
        >>>
        >>> neph.evaluate(d, rh=85.)

        """
        total_cscat = self._sum_across_distribution(distribution, rh=rh, **kwargs)

        pm1 = total_cscat / self.pm1_ratio
        pm25 = total_cscat / self.pm25_ratio
        pm10 = total_cscat / self.pm10_ratio

        return total_cscat, pm1, pm25, pm10


__all__ = [
    'OPC',
    'Nephelometer'
]



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['opcsim', 'opcsim.equations']

package_data = \
{'': ['*']}

install_requires = \
['matplotlib>=3.2.1,<4.0.0',
 'numpy>=1.18.4,<2.0.0',
 'pandas>=1.0.3,<2.0.0',
 'scipy>=1.4.1,<2.0.0',
 'seaborn>=0.10.1,<0.11.0']

setup_kwargs = {
    'name': 'opcsim',
    'version': '0.5.0a4',
    'description': 'OPCSIM: simulating low-cost optical particle counters',
    'long_description': '[![PyPI version](https://badge.fury.io/py/opcsim.svg)](https://badge.fury.io/py/opcsim)\n[![license](https://img.shields.io/github/license/mashape/apistatus.svg)](https://github.com/dhhagan/opcsim/blob/master/LICENSE)\n![run and build](https://github.com/dhhagan/opcsim/workflows/run%20and%20build/badge.svg)\n[![codecov](https://codecov.io/gh/dhhagan/opcsim/branch/master/graph/badge.svg)](https://codecov.io/gh/dhhagan/opcsim)\n![Docker Pulls](https://img.shields.io/docker/pulls/dhhagan/opcsim)\n![Docker Stars](https://img.shields.io/docker/stars/dhhagan/opcsim)\n\n# opcsim\n\nopcsim is a Python library for simulating low-cost Optical Particle Sensors (both Optical Particle Counters and Nephelometers) and\ntheir response to various aerosol distributions.\n\n## Documentation\n\nFull online documentation can be found [here][1].\n\nThe docs include a [tutorial][2], an [example gallery][3], and an [API Reference][4].\n\nIn addition, documentation can be built locally for development purposes. To do so, please check out the complete details in the *contributing to opcsim* section of the documentation.\n\n## Docker\n\nIf you are familiar with Docker, there is a Docker image available to get up and running with OPCSIM with ease. To get started \nwith an ephemeral container with a jupyter lab interface, navigate to your preferred working directory and execute:\n\n```sh\n$ docker run --rm -p 8888:8888 -e JUPYTER_ENABLE_LAB=yes -v "$PWD":/home/joyvan/work dhhagan/opcsim:latest\n```\n\nOnce executed, you should see the url with token in your terminal that will allow you to bring up the jupyter lab instance.\n\n\n## Dependencies\n\nOpcsim is supported for python3.6.1+.\n\nInstallation requires [scipy][5], [numpy][6], [pandas][7], [matplotlib][8],\nand [seaborn][9].\n\n\n## Installation\n\nTo install (or upgrade to) the latest stable release:\n\n```sh\n\n$ pip install opcsim [--upgrade]\n```\n\nTo install the development version directly from GitHub using pip:\n\n```sh\n\n$ pip install git+https://github.com/dhhagan/opcsim.git\n```\n\nIn addition, you can either clone the repository and install from source or download/unzip the zip file and install from source using poetry:\n\n```sh\n\n$ git clone https://github.com/dhhagan/opcsim.git\n$ cd /opcsim\n$ poetry install\n```\n\n## Testing\n\nAll tests are automagically run via GitHub actions and Travis.ci. For results of these tests, please click on the link in the above travis badge. In addition, you can run tests locally using poetry.\n\nTo run tests locally:\n\n```sh\n\n$ poetry run pytest tests\n```\n\n\n## Development\n\n**opcsim** development takes place on GitHub. Issues and bugs can be submitted and tracked via the [GitHub Issue Tracker][10] for this repository. As of `v0.5.0`, *opcsim* uses [poetry][11] for versioning and managing dependencies and releases.\n\n\n[1]: https://dhhagan.github.io/opcsim/\n[2]: https://dhhagan.github.io/opcsim/tutorial.html\n[3]: https://dhhagan.github.io/opcsim/examples/index.html\n[4]: https://dhhagan.github.io/opcsim/api.html\n[5]: https://www.scipy.org/\n[6]: http://www.numpy.org/\n[7]: http://pandas.pydata.org/\n[8]: http://matplotlib.org/\n[9]: https://seaborn.pydata.org/\n[10]: https://github.com/dhhagan/opcsim/issues\n[11]: https://python-poetry.org/\n',
    'author': 'David H Hagan',
    'author_email': 'dhagan@mit.edu',
    'maintainer': 'David H Hagan',
    'maintainer_email': 'dhagan@mit.edu',
    'url': 'https://dhhagan.github.io/opcsim/',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.6.1,<4.0.0',
}


setup(**setup_kwargs)
